/* adal.c */

#include <Pilot.h>

#include <Graffiti.h>		/* for GrfSetState() */
#include <Hardware.h>		/* for hwrLaunchBtnLeft, etc. */
#include <SysEvtMgr.h>		/* for EvtFlushNextPenStroke() */

#include "adal.h"
#include "adal_rcp.h"


/*
 * macros
 */

#if defined(__GNUC__) && defined(USE_A4_REGISTER)
#define ARG(arg)
#define ARG_(arg)
#define SAVE_A4_REGISTER void *old_g = g;
#define RESTORE_A4_REGISTER g = old_g;
#else
#define ARG(arg) arg
#define ARG_(arg) arg,
#define SAVE_A4_REGISTER
#define RESTORE_A4_REGISTER
#endif

#ifdef __GNUC__
register void *reg_a7 asm("%sp");
#define __getA7() (long) reg_a7
#define __setA7(a7) reg_a7 = (void *) (a7)
#else /* __MWERKS__ */
long __getA7(void) = { 0x200F };
void __setA7(long a7:__d0) = { 0x2E40 };
#endif

#define printable(c) ((c) >= ' ' && (c) <= 'z')


/*
 * constants
 */

#define HACK_DB_TYPE 'HACK'

#define DEFAULT_CARD_NO 0	/* zzz */
#define HISTORY_DB_NAME "App/DA Launcher History"

#define HISTORY_RSRC_TYPE 'HIST'
#define HISTORY_RSRC_ID 1000
#define HISTORY_RSRC_SIZE (sizeof(struct history_record) \
	    + sizeof(struct history_entry) * HISTORY_MAX_COUNT)

#define FTR_ORIG 1000
#define FTR_GLOBALS 9000

#define APP_DB_TYPE 'appl'
#define PANEL_DB_TYPE 'panl'
#define PREFS_CREATOR 'pref'

#define DA_DB_TYPE 'DAcc'
#define DA_RSRC_TYPE 'code'
#define DA_RSRC_ID 1000

#define MIN_STACK_SIZE 256

#define KEYED_DELAY 30

#define LIST_HEIGHT 14
#define LIST_ITEM_HEIGHT 11
#define LIST_LEFT 2
#define LIST_BOTTOM 158

#define SAVE_WIDTH 80
#define SAVE_HEIGHT 160

enum ret_code {
    DO_NOTHING,
    DO_LAUNCH,
    SWITCH_MODE
};


/*
 * types
 */

#define DB_INFO_LENGTH 4

struct db_info {
    UInt card_no;
    LocalID db_id;
    union {
	long l;
	char s[DB_INFO_LENGTH];
    } name;
};

struct globals {
    ULong creator;
    struct adal_prefs prefs;
    int mode;
    int new_mode;
    FormPtr saved_form;
    Word saved_form_id;
    WinHandle saved_active_win;
    WinHandle saved_draw_win;
    FontID saved_font;
    FormPtr form;
    ListPtr list;
    struct db_info *db_list;
    int db_cnt;
    DmOpenRef history_db;
    VoidHand history_rsrc;
    struct history_record *history_record;
    UInt card_no;
    LocalID db_id;
    char name[dmDBNameLength];
    char keyed_name[dmDBNameLength];
    int keyed_length;
    int keyed_first;
    int keyed_last;
};

#if defined(__GNUC__) && defined(USE_A4_REGISTER)
register struct globals *g asm("%a4");
#endif


/*
 * functions
 */

static Boolean PointInRect(PointType *p, int x, int y, int width, int height)
{
    if (p->x >= x && p->x < x + width
     && p->y >= y && p->y < y + height) {
	return true;
    } else {
	return false;
    }
}

static int find_part(PointType *start, PointType *end)
{
    struct adal_prefs prefs;

    if (PointInRect(start, hwrLaunchBtnLeft, hwrLaunchBtnTop,
	    hwrLaunchBtnWidth, hwrLaunchBtnHeight)
     && PointInRect(end, hwrLaunchBtnLeft, hwrLaunchBtnTop,
	    hwrLaunchBtnWidth, hwrLaunchBtnHeight)) {

	get_prefs(&prefs);

	if (start->x < hwrLaunchBtnLeft + hwrLaunchBtnWidth / 2) {
	    if (start->y < hwrLaunchBtnTop + hwrLaunchBtnHeight / 2) {
		return prefs.button1;
	    } else {
		return prefs.button2;
	    }
	} else {
	    if (start->y < hwrLaunchBtnTop + hwrLaunchBtnHeight / 2) {
		return prefs.button3;
	    } else {
		return prefs.button4;
	    }
	}
    }

    if (PointInRect(start, hwrLaunchBtnLeft, hwrLaunchBtnTop,
	    hwrLaunchBtnWidth, hwrLaunchBtnHeight)
     && PointInRect(end, hwrMenuBtnLeft, hwrMenuBtnTop,
	    hwrMenuBtnWidth, hwrMenuBtnHeight)) {
	return MODE_SWITCH_BACK;
    }

    return MODE_NONE;
}

static Boolean IsDeviceLocked()
{
    DWord rom_version;
    SystemPreferencesTypeV10 sys_prefs;

    if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != noErr) {
	return false;
    } else if (sysGetROMVerMajor(rom_version) < 2) {
	PrefGetPreferences((SystemPreferencesPtr) &sys_prefs);
	return sys_prefs.deviceLocked;
    } else {
	return PrefGetPreference(prefDeviceLocked);
    }
}

static ULong GetCurAppCreator()
{
    UInt card_no;
    LocalID db_id;
    ULong type, creator;

    if (SysCurAppDatabase(&card_no, &db_id) == noErr
     && DmDatabaseInfo(card_no, db_id, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, NULL, NULL, &type, &creator) == noErr) {
	if (type == PANEL_DB_TYPE) {
	    return PREFS_CREATOR;
	} else {
	    return creator;
	}
    }

    return 0;
}

static char to_lower(char c)
{
    return (c >= 'A' && c <= 'Z') ? (c - 'A' + 'a') : c;
}

static Err get_db_name(ARG_(struct globals *g) int i)
{
    if (g->mode == MODE_HISTORY) {
	StrCopy(g->name, g->history_record->entries[i].name);
	return noErr;
    } else {
	return DmDatabaseInfo(
		g->db_list[i].card_no, g->db_list[i].db_id, g->name,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }
}


/*
 * db_list management
 */

static void put_db_info(ARG(struct globals *g))
{
    char *name = g->keyed_name;
    UInt attr;
    struct db_info db_info;
    int j;

    if (DmDatabaseInfo(g->card_no, g->db_id, name, &attr,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) == noErr
     && (attr & dmHdrAttrResDB) != 0) {
	db_info.card_no = g->card_no;
	db_info.db_id = g->db_id;
	db_info.name.l = 0;
	for (j = 0; j < DB_INFO_LENGTH && name[j] != '\0'; ++j) {
	    db_info.name.s[j] = to_lower(name[j]);
	}
	for (j = g->db_cnt - 1; j >= 0; --j) {
	    if (db_info.name.l < g->db_list[j].name.l
	     || (get_db_name(ARG_(g) j) == noErr
	      && StrCaselessCompare(name, g->name) < 0)) {
		g->db_list[j + 1] = g->db_list[j];
	    } else {
		break;
	    }
	}
	g->db_list[j + 1] = db_info;
	++g->db_cnt;
    }
}

static void gen_db_list(ARG(struct globals *g))
{
    Word mode;
    ULong type;
    Boolean new_search;
    DmSearchStateType state;
    int cnt;

    cnt = 0;
    for (mode = g->mode & MODE_BOTH, type = APP_DB_TYPE;
	 mode != MODE_NONE; mode >>= 1, type = DA_DB_TYPE) {
	if (mode & 1) {
	    for (new_search = true;
		 DmGetNextDatabaseByTypeCreator(new_search, &state,
			type, 0, false, &g->card_no, &g->db_id) == noErr;
		 new_search = false) {
		++cnt;
	    }
	}
    }

    if (cnt == 0
     || (g->db_list = MemPtrNew(sizeof(struct db_info) * cnt)) == NULL) {
	return;
    }

    for (mode = g->mode & MODE_BOTH, type = APP_DB_TYPE;
	 mode != MODE_NONE; mode >>= 1, type = DA_DB_TYPE) {
	if (mode & 1) {
	    for (new_search = true;
		 g->db_cnt < cnt
	      && DmGetNextDatabaseByTypeCreator(new_search, &state,
			type, 0, false, &g->card_no, &g->db_id) == noErr;
		 new_search = false) {
		put_db_info(ARG(g));
	    }
	}
    }
}

static Boolean open_history_db(ARG(struct globals *g))
{
    Boolean res_db;
    struct history_record *p;

    g->history_rsrc = NULL;
    g->history_record = NULL;

    g->history_db = DmOpenDatabaseByTypeCreator(
	    HISTORY_DB_TYPE, CREATOR, dmModeReadOnly);
    if (g->history_db != NULL) {
	if (DmOpenDatabaseInfo(
		g->history_db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false
	 && (g->history_rsrc = DmGet1Resource(
		HISTORY_RSRC_TYPE, HISTORY_RSRC_ID)) != NULL) {
	    if (MemHandleSize(g->history_rsrc) == HISTORY_RSRC_SIZE
	     && (g->history_record = p = MemHandleLock(g->history_rsrc))
		    != NULL) {
		if (p->entry_size == sizeof(struct history_entry)
		 && p->entry_cnt > 0 && p->entry_cnt <= HISTORY_MAX_COUNT) {
		    return true;
		}
	    }
	}
    }

    return false;
}

static void gen_history_list(ARG(struct globals *g))
{
    char *name;
    int cnt, i, j;

    if (open_history_db(ARG(g))) {
	cnt = min(g->history_record->entry_cnt, DEFAULT_HISTORY_COUNT);
	g->db_list = MemPtrNew(sizeof(struct db_info) * cnt);
	if (g->db_list != NULL) {
	    for (i = 0; i < cnt; ++i) {
		g->db_list[i].name.l = 0;
		name = g->history_record->entries[i].name;
		for (j = 0; j < DB_INFO_LENGTH && name[j] != '\0'; ++j) {
		    g->db_list[i].name.s[j] = to_lower(name[j]);
		}
	    }
	    g->db_cnt = cnt;
	}
    }
}

static void close_history_db(ARG(struct globals *g))
{
    if (g->history_db != NULL) {
	if (g->history_rsrc != NULL) {
	    if (g->history_record != NULL) {
		MemHandleUnlock(g->history_rsrc);
	    }
	    DmReleaseResource(g->history_rsrc);
	}
	DmCloseDatabase(g->history_db);
    }
}

static void free_db_list(ARG(struct globals *g))
{
    if (g->db_list != NULL) {
	MemPtrFree(g->db_list);
    }

    if (g->mode == MODE_HISTORY) {
	close_history_db(ARG(g));
    }
}


/*
 * user interaction
 */

static void draw_list_item(Word item_no, RectanglePtr bounds, CharPtr *dummy)
{
    struct globals ARG_(*g) *gg;
    Int width, length;
    Boolean fits;

    SAVE_A4_REGISTER

    if (FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &gg) == noErr
     && (g = gg) != NULL && g->creator == CREATOR
     && item_no >= 0 && item_no < g->db_cnt
     && get_db_name(ARG_(g) item_no) == noErr) {
	width = bounds->extent.x;
	length = StrLen(g->name);
	FntCharsInWidth(g->name, &width, &length, &fits);
	WinDrawChars(g->name, length, bounds->topLeft.x, bounds->topLeft.y);
    } else {
	WinDrawChars("none", 4, bounds->topLeft.x, bounds->topLeft.y);
    }

    RESTORE_A4_REGISTER
}

static int match_name(ARG_(struct globals *g) int i, int *jp)
{
    int j;
    char c1, c2;

    for (j = 0; j < g->keyed_length; ++j) {
	c1 = g->keyed_name[j];
	if (j < DB_INFO_LENGTH) {
	    c2 = g->db_list[i].name.s[j];
	} else {
	    if (j == DB_INFO_LENGTH) {
		if (get_db_name(ARG_(g) i) != noErr) {
		    *jp = j;
		    return c1;
		}
	    }
	    c2 = to_lower(g->name[j]);
	}
	if (c1 != c2) {
	    *jp = j;
	    return c1 - c2;
	}
    }
    return 0;
}

static Boolean match_name2(ARG_(struct globals *g) int i, int j_max)
{
    int j, k;
    char c1, c2;

    if (get_db_name(ARG_(g) i) != noErr) {
	return false;
    }

    for (j = 0, k = 0; j < g->keyed_length; ++j, ++k) {
	c1 = g->keyed_name[j];
	c2 = to_lower(g->name[k]);
	if (j < j_max) {
	    if (c1 != c2) { return false; }
	} else {
	    while (c1 != c2) {
		if (c2 == '\0') { return false; }
		c2 = to_lower(g->name[++k]);
	    }
	}
    }
    return true;
}

static void find_db(ARG(struct globals *g))
{
    int first, last, i, j, j_max, j_first, j_last, match;

    first = -1;
    last = 0;

    j_max = 0;
    j_first = 0;
    j_last = -1;

    for (i = 0; i < g->db_cnt; ++i) {
	match = match_name(ARG_(g) i, &j);
	if (match == 0) {
	    if (first < 0) { first = i; }
	    last = i;
	} else if (j >= j_max) {
	    if (j > j_max) {
		j_max = j;
		j_first = i;
	    }
	    j_last = i;
	} else if (match < 0 && g->mode != MODE_HISTORY) {
	    break;
	}
    }

    if (first < 0) {
	for (i = j_first; i <= j_last; ++i) {
	    if (match_name2(ARG_(g) i, j_max)) {
		if (first < 0) { first = i; }
		last = i;
	    }
	}
    }

    g->keyed_first = first;
    g->keyed_last = last;
}

static short event_loop(ARG(struct globals *g))
{
    EventType event;
    Word chr;
    ListPtr list = g->list;
    short new_sel;

    g->keyed_length = 0;

    while (g->new_mode == MODE_NONE) {
	EvtGetEvent(&event, evtWaitForever);

	chr = (event.eType == keyDownEvent) ? event.data.keyDown.chr : 0;
	if (event.eType == appStopEvent
	 || chr == menuChr || chr == launchChr
	 || chr == findChr || ChrIsHardKey(chr)) {
	    EvtAddEventToQueue(&event);
	    break;
	} else if (chr == backspaceChr) {
	    if (g->prefs.backspace == BACKSPACE_PREVIOUS_ITEM) {
		new_sel = LstGetSelection(list) - 1;
		if (new_sel < 0) { new_sel = 0; }
		LstSetSelection(list, new_sel);
	    } else if (g->prefs.backspace == BACKSPACE_CLEAR_CHAR
		    && g->keyed_length >= 2) {
		--g->keyed_length;
		goto find;		/* !!! */
	    } else {
		g->keyed_length = 0;
		LstSetSelection(list, (g->db_cnt > 0) ? 0 : noListSelection);
	    }
	} else if (chr == linefeedChr) {
	    return LstGetSelection(list);
	} else if (chr == pageUpChr) {
	    new_sel = LstGetSelection(list) -
		((g->prefs.page_key == PAGEKEY_SCROLL_LINE) ? 1 : LIST_HEIGHT);
	    if (new_sel < 0) { new_sel = 0; }
	    LstSetSelection(list, new_sel);
	} else if (chr == pageDownChr) {
	    new_sel = LstGetSelection(list) +
		((g->prefs.page_key == PAGEKEY_SCROLL_LINE) ? 1 : LIST_HEIGHT);
	    if (new_sel >= g->db_cnt) { new_sel = g->db_cnt - 1; }
	    LstSetSelection(list, new_sel);
	} else if (chr == spaceChr && g->prefs.space == SPACE_NEXT_ITEM) {
	    new_sel = LstGetSelection(list) + 1;
	    if (new_sel >= g->db_cnt) { new_sel = g->db_cnt - 1; }
	    LstSetSelection(list, new_sel);
	} else if (printable(chr) && g->keyed_length < sizeof g->keyed_name) {
	    g->keyed_name[g->keyed_length++] = to_lower(chr);
	find:
	    find_db(ARG(g));
	    if (g->keyed_first >= 0) {
		if (g->prefs.input_feedback) {
		    LstSetSelection(list, g->keyed_last);
		    SysTaskDelay(KEYED_DELAY);
		}
		if (g->keyed_first == g->keyed_last) {
		    return g->keyed_first;
		} else {
		    LstSetSelection(list, g->keyed_first);
		}
	    }
	} else {
	    if (! SysHandleEvent(&event)
	     && ! FrmHandleEvent(g->form, &event)) {
		if (event.eType == lstSelectEvent) {
		    return event.data.lstSelect.selection;
		} else if (event.eType == penDownEvent) {
		    break;
		}
	    }
	}

	if (g->new_mode == g->mode) {
	    if (g->prefs.tap_again == TAPAGAIN_DO_NOTHING) {
		g->new_mode = MODE_NONE;
	    } else if (g->prefs.tap_again == TAPAGAIN_CLEAR_INPUT) {
		g->keyed_length = 0;
		LstSetSelection(list, (g->db_cnt > 0) ? 0 : noListSelection);
		g->new_mode = MODE_NONE;
	    } else if (g->prefs.tap_again == TAPAGAIN_OS_LAUNCHER) {
		event.eType = keyDownEvent;
		event.data.keyDown.chr = launchChr;
		event.data.keyDown.keyCode = 0;
		event.data.keyDown.modifiers = commandKeyMask;
		EvtAddEventToQueue(&event);
	    }
	}
    }

    return noListSelection;
}

static int open_launcher(ARG(struct globals *g))
{
    DmOpenRef db;
    Boolean res_db;
    WinHandle save;
    RectangleType r;
    Err err;
    ListPtr list;
    int list_cnt, list_height;
    short sel;
    DmSearchStateType state;
    int ret_code = DO_NOTHING;

    db = DmOpenDatabaseByTypeCreator(HACK_DB_TYPE, CREATOR, dmModeReadOnly);
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false) {
	    g->form = FrmInitForm(MAIN_FORM_ID);
	    FrmSetActiveForm(g->form);

	    GrfSetState(false, false, false);

	    RctSetRectangle(&r, 0, 0, SAVE_WIDTH, SAVE_HEIGHT);
	    save = WinSaveBits(&r, &err);

	    g->db_list = NULL;
	    g->db_cnt = 0;
	    if (g->mode == MODE_HISTORY) {
		gen_history_list(ARG(g));
	    } else {
		gen_db_list(ARG(g));
	    }

	    g->list = list = FrmGetObjectPtr(
		    g->form, FrmGetObjectIndex(g->form, MAIN_LIST_ID));
	    LstSetDrawFunction(list, &draw_list_item);
	    list_cnt = (g->db_cnt > 0) ? g->db_cnt : 1;
	    LstSetListChoices(list, NULL, list_cnt);
	    list_height = (list_cnt > LIST_HEIGHT) ? LIST_HEIGHT : list_cnt;
	    LstSetPosition(list,
		    LIST_LEFT, LIST_BOTTOM - LIST_ITEM_HEIGHT * list_height);
	    LstSetHeight(list, list_height);
	    if (g->db_cnt == 0) {
		LstSetSelection(list, noListSelection);
	    }
	    LstEraseList(list);
	    LstDrawList(list);

	    sel = event_loop(ARG(g));

	    if (sel >= 0 && sel < g->db_cnt) {
		if (g->mode == MODE_HISTORY) {
		    if (DmGetNextDatabaseByTypeCreator(true, &state,
			    g->history_record->entries[sel].type,
			    g->history_record->entries[sel].creator,
			    true, &g->card_no, &g->db_id) == noErr) {
			ret_code = DO_LAUNCH;
		    }
		} else {
		    g->card_no = g->db_list[sel].card_no;
		    g->db_id = g->db_list[sel].db_id;
		    ret_code = DO_LAUNCH;
		}
	    }

	    if (g->new_mode != MODE_NONE) {
		if (g->new_mode == g->mode) {
		    ret_code = DO_NOTHING;
		} else {
		    ret_code = SWITCH_MODE;
		}
	    }

	    free_db_list(ARG(g));

	    WinRestoreBits(save, 0, 0);

	    FrmDeleteForm(g->form);
	}
	DmCloseDatabase(db);
    }

    return ret_code;
}


/*
 * launching
 */

static void record_history(ARG_(struct globals *g) ULong type, ULong creator)
{
    DmOpenRef db;
    Boolean res_db;
    VoidHand h;
    struct history_record *p;
    struct history_record record;
    struct history_entry entry;
    int i;

    entry.type = type;
    entry.creator = creator;
    MemSet(entry.name, sizeof entry.name, 0);
    StrCopy(entry.name, g->name);

    db = DmOpenDatabaseByTypeCreator(HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
    if (db == NULL) {
	if (DmCreateDatabase(DEFAULT_CARD_NO, HISTORY_DB_NAME,
		CREATOR, HISTORY_DB_TYPE, true) == noErr) {
	    db = DmOpenDatabaseByTypeCreator(
		    HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
	}
    }
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false) {
	    h = DmGet1Resource(HISTORY_RSRC_TYPE, HISTORY_RSRC_ID);
	    if (h == NULL) {
		h = DmNewResource(db,
			HISTORY_RSRC_TYPE, HISTORY_RSRC_ID, HISTORY_RSRC_SIZE);
		if (h != NULL
		 && (p = MemHandleLock(h)) != NULL) {
		    DmSet(p, 0, HISTORY_RSRC_SIZE, 0);
		    record.entry_size = sizeof(struct history_entry);
		    record.entry_cnt = 0;
		    DmWrite(p, 0, &record, sizeof record);
		    MemHandleUnlock(h);
		}
	    }
	    if (h != NULL) {
		if (MemHandleSize(h) == HISTORY_RSRC_SIZE
		 && (p = MemHandleLock(h)) != NULL) {
		    if (p->entry_size == sizeof(struct history_entry)
		     && p->entry_cnt >= 0
		     && p->entry_cnt <= HISTORY_MAX_COUNT) {
			for (i = 0; i < p->entry_cnt; ++i) {
			    if (p->entries[i].creator == entry.creator
			     && p->entries[i].type == entry.type) {
				break;
			    }
			}
			if (i == p->entry_cnt) {
			    if (i >= HISTORY_MAX_COUNT) {
				i = HISTORY_MAX_COUNT - 1;
			    } else {
				record = *p;
				++record.entry_cnt;
				DmWrite(p, 0, &record, sizeof record);
			    }
			}
			for ( ; i > 0; --i) {
			    DmWrite(p, sizeof(struct history_record)
				     + sizeof(struct history_entry) * i,
				    &p->entries[i - 1],
				    sizeof(struct history_entry));
			}
			DmWrite(p, sizeof(struct history_record),
				&entry, sizeof entry);
		    }
		    MemHandleUnlock(h);
		}
		DmReleaseResource(h);
	    }
	}
	DmCloseDatabase(db);
    }
}

static void launch_da(UInt card_no, LocalID db_id, ULong stack_size)
{
    DmOpenRef db;
    Boolean res_db;
    VoidHand h;
    void (*da_entry)();
    VoidPtr new_stack;
    long saved_stack;

    db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false
	 && (h = DmGet1Resource(DA_RSRC_TYPE, DA_RSRC_ID)) != NULL) {
	    da_entry = MemHandleLock(h);
	    if (da_entry != NULL) {
		new_stack = NULL;
		stack_size &= ~1L;
		if (stack_size >= MIN_STACK_SIZE) {
		    new_stack = MemPtrNew(stack_size);
		}
		saved_stack = __getA7();	/* to shut up gcc -W ... */
		if (new_stack != NULL) {
		    __setA7((long) new_stack + stack_size);
		}
		InsPtEnable(false);
		(*da_entry)();
		if (new_stack != NULL) {
		    __setA7(saved_stack);
		    MemPtrFree(new_stack);
		}
		MemHandleUnlock(h);
	    }
	    DmReleaseResource(h);
	}
	DmCloseDatabase(db);
    }
}

static void do_launch(ARG(struct globals *g))
{
    UInt attr;
    ULong type;
    ULong creator;

    if (DmDatabaseInfo(g->card_no, g->db_id, g->name, &attr,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator) == noErr
     && (attr & dmHdrAttrResDB) != 0) {
	if (g->prefs.record_history) {
	    record_history(ARG_(g) type, creator);
	}
	if (type == APP_DB_TYPE) {
	    SysUIAppSwitch(
		    g->card_no, g->db_id, sysAppLaunchCmdNormalLaunch, 0);
	} else if (type == DA_DB_TYPE) {
	    launch_da(g->card_no, g->db_id, g->prefs.stack_size);

	    if (g->saved_form_id != 0) {
		FrmSetActiveForm(g->saved_form);
	    }
	    WinSetActiveWindow(g->saved_active_win);
	    WinSetDrawWindow(g->saved_draw_win);
	}
    }
}

static void switch_back()
{
    ARG(struct globals *g;)
    ULong cur_app;
    DmSearchStateType state;
    int i, ret_code = DO_NOTHING;

    SAVE_A4_REGISTER

    if (IsDeviceLocked() == false) {
	g = MemPtrNew(sizeof(struct globals));
	if (g != NULL) {
	    get_prefs(&g->prefs);

	    cur_app = GetCurAppCreator();

	    if (open_history_db(ARG(g))) {
		for (i = 0; i < g->history_record->entry_cnt; ++i) {
		    if (g->history_record->entries[i].type == APP_DB_TYPE
		     && g->history_record->entries[i].creator != cur_app
		     && DmGetNextDatabaseByTypeCreator(true, &state,
			    g->history_record->entries[i].type,
			    g->history_record->entries[i].creator,
			    true, &g->card_no, &g->db_id) == noErr) {
			ret_code = DO_LAUNCH;
			break;
		    }
		}
	    }

	    close_history_db(ARG(g));

	    if (ret_code == DO_LAUNCH) {
		do_launch(ARG(g));
	    }

	    MemPtrFree(g);
	}
    }

    RESTORE_A4_REGISTER
}


/*
 * main routines
 */

static void do_launcher(int part)
{
    struct globals ARG_(*g) *gg;
    Err err;
    int ret_code;

    SAVE_A4_REGISTER

    err = FtrGet(CREATOR, FTR_GLOBALS, (DWord *) &gg);
    if (err == noErr) {
	g = gg;
	if (g != NULL && g->creator == CREATOR) {
	    g->new_mode = part;
	}
    } else if (err == ftrErrNoSuchFeature
	    && IsDeviceLocked() == false) {
	g = MemPtrNew(sizeof(struct globals));
	if (g != NULL) {
	    if (FtrSet(CREATOR, FTR_GLOBALS, (DWord) g) == noErr) {
		g->creator = CREATOR;
		get_prefs(&g->prefs);

		GsiSetShiftState(0, 0);

		g->saved_form = FrmGetActiveForm();
		g->saved_form_id = FrmGetActiveFormID();
		g->saved_active_win = WinGetActiveWindow();
		g->saved_draw_win = WinGetDrawWindow();
		g->saved_font = FntSetFont(stdFont);

		do {
		    g->mode = part;
		    g->new_mode = MODE_NONE;
		    ret_code = open_launcher(ARG(g));
		    part = g->new_mode;
		} while (ret_code == SWITCH_MODE);

		if (g->saved_form_id != 0) {
		    FrmSetActiveForm(g->saved_form);
		}
		WinSetActiveWindow(g->saved_active_win);
		WinSetDrawWindow(g->saved_draw_win);
		FntSetFont(g->saved_font);

		FtrUnregister(CREATOR, FTR_GLOBALS);

		if (ret_code == DO_LAUNCH) {
		    do_launch(ARG(g));
		}
	    }
	    MemPtrFree(g);
	}
    }

    RESTORE_A4_REGISTER
}

Err MyHandler(PointType *start, PointType *end)
{
    int part;
    Err (*orig_trap)(PointType *start, PointType *end);

    part = find_part(start, end);
    if (part == MODE_APP || part == MODE_DA || part == MODE_BOTH
     || part == MODE_HISTORY) {
	EvtFlushNextPenStroke();
	do_launcher(part);
    } else if (part == MODE_SWITCH_BACK) {
	EvtFlushNextPenStroke();
	switch_back();
    } else if (FtrGet(CREATOR, FTR_ORIG, (DWord *) &orig_trap) == noErr) {
	return (*orig_trap)(start, end);
    }

    return noErr;
}
